"""Utilities for tracking resources"""
from threading import Semaphore, Lock, Event, Thread
from typing import List, Dict, Optional, Union
from time import monotonic
from math import inf
import logging

logger = logging.getLogger(__name__)

_CANCEL_CHECK_FREQ = 1.0


class ResourceCounter:
    """Utility class for keeping track of resources available for different tasks.

    The class manages two pieces of state: the amount of resources allocated to a certain task,
    and the amount of resources that are currently available for that task.
    Users of this class can change either state using a series of thread-safe methods.

    *Tracking Allocations*: The resource counter is initialized with a certain count of resources,
    which represent the total number of a certain computing device available (e.g., node, GPU).
    They all begin as "unallocated" for any task.

    Users change the amount of resources dedicated to tasks by "reallocating" them from one task to another.
    The :meth:`reallocate` method achieves this by requesting a certain number of units from one task
    and adding them to a second task's available resources once those units are marked as available.

    *Tracking Utilization*: The amount of resources in use for a certain task is tracked by an internal counter.
    Users of this class request the use a certain number of resources by calling the :meth:`acquire` method.
    The method blocks until either the request is completely fulfilled (i.e., the specified amount of resources
    are marked as available) or the operation times out.

    Resources are marked as available again using the :meth:`release` method.
    The release method marks those resources as available to be re-used for other tasks of the same type.
    Resources must be re-allocated using :meth:`reallocate`.

    **Implementation**: All of the operations described above are thread-safe.
    Resource utilization is tracked using a semaphore so that threads can acquire and release resources simultaneously.
    Resources are acquired as first-come-first-served by using a lock to control access to the "acquire" function
    of the resource utilization semaphore.
    """

    def __init__(self, total_slots: int, task_types: List[str]):
        """
        Args:
            total_slots: Total number of nodes available to the resources
            task_types: Names of task types
        """
        # Save the total number of nodes available
        self._total_slots = total_slots

        # Add a "global" task type
        my_tasks: List[Optional[str]] = task_types.copy()
        if None in my_tasks:
            raise ValueError("`None` is reserved as the global task name")
        my_tasks.append(None)

        # Create semaphores that track the resources available tasks
        self._availability: Dict[Optional[str], Semaphore] = dict(
            (t, Semaphore(value=0)) for t in my_tasks
        )
        self._availability_lock: Dict[Optional[str], Lock] = dict(
            (t, Lock()) for t in my_tasks
        )

        # Create counters to represent the amount of resources allocated to each task
        self._allocation: Dict[Optional[str], int] = dict(
            (t, 0) for t in my_tasks
        )
        self._allocation_lock: Dict[Optional[str], Lock] = dict(
            (t, Lock()) for t in my_tasks
        )

        # Mark the number of unallocated slots
        for _ in range(self._total_slots):
            self._availability[None].release()
        self._allocation[None] = self._total_slots

        logger.info(f'Created a resource tracker with {total_slots} slots and {len(task_types)} task types: {", ".join(task_types)}')

    @property
    def unallocated_slots(self) -> int:
        """Number of unallocated slots"""
        return self._allocation[None]

    def allocated_slots(self, task: str) -> int:
        """Number of slots allocated to a certain task

        Args:
            task: Name of the task
        """
        if task not in self._availability:
            raise KeyError(f'Unknown task name: {task}')
        return self._allocation[task]

    def available_slots(self, task: Optional[str]) -> int:
        """Get the number of nodes available for a certain task

        Args:
            task: Name of the task
        Returns:
            Number of slots available for that task
        """
        if task not in self._availability:
            raise KeyError(f'Unknown task name: {task}')
        return self._availability[task]._value

    def release(self, task: Optional[str], n_slots: int, rerequest: bool = False, timeout: float = -1) -> Optional[bool]:
        """Register that nodes for a particular task are available
        and, optionally, re-request those nodes for the same task.

        Blocks until the task request completes

        Args:
            task: Name of the task
            n_slots: Number of slots to mark as available
            rerequest: Whether to re-request nodes immediately after releasing them
            timeout: Maximum time to wait for the request to be filled
        Returns:
            Whether the re-request was fulfilled
        """

        for _ in range(n_slots):
            self._availability[task].release()  # TODO (wardlt): Py3.9 lets you release counter by >1
        if rerequest:
            return self.acquire(task, n_slots, timeout=timeout)
        return None

    def acquire(self, task: Optional[str], n_slots: int, timeout: float = -1., cancel_if: Optional[Event] = None) -> bool:
        """Request a certain number of nodes for a particular task

        Draws only from the pool of nodes allocated to this task

        Args:
            task: Name of the task
            n_slots: Number of slots to request
            timeout: Maximum time to wait for the request to be filled
            cancel_if: Cancel the request if this event happens
        Returns:
            Whether the request was fulfilled
        """

        # Determine when this operation will time out
        if timeout > 0:
            end_time = monotonic() + timeout
        else:
            end_time = inf

        # Set a small timeout if we are also checking for a condition
        if cancel_if:
            timeout = _CANCEL_CHECK_FREQ

        # Acquire the lock for getting the hold
        lock_acquired = False
        while not lock_acquired:
            lock_acquired = self._availability_lock[task].acquire(timeout=timeout)
            if (cancel_if is not None and cancel_if.is_set()) or monotonic() > end_time:
                if lock_acquired:
                    self._availability_lock[task].release()
                return False

        # Wait until all nodes are acquired
        n_acquired = 0
        success = True
        while n_acquired < n_slots:
            acq_success = self._availability[task].acquire(timeout=timeout if timeout > 0 else None)
            if (cancel_if is not None and cancel_if.is_set()) or monotonic() > end_time:
                success = False
                break
            if acq_success:
                n_acquired += 1

        # Let another thread use this class
        self._availability_lock[task].release()

        # If you were not successful, give the resources back
        if not success:
            for _ in range(n_acquired):
                self._availability[task].release()

        return success

    def reallocate(self, task_from: Optional[str], task_to: Optional[str], n_slots: Union[int, str],
                   timeout: float = -1, cancel_if: Optional[Event] = None) -> bool:
        """Transfer computer resources from one task to another

        Args:
            task_from: Which task to pull resources from (None to request un-allocated nodes)
            task_to: Which task to add resources to (None to de-allocate nodes)
            n_slots: Number of nodes to request. Set to "all" to reallocate all slots (all allocated slots, not just all available slots)
            timeout: Maximum time to wait for the request to be filled
            cancel_if: Cancel the request if this event happens
        Returns:
            Whether request was fulfilled
        """
        if task_to == task_from:
            raise ValueError(f'Resources cannot be moved between the same pool. task_from = "{task_from}" = task_to')

        # Pull nodes from the remaining
        with self._allocation_lock[task_from]:
            if n_slots == "all":
                n_slots = self.allocated_slots(task_from)
            acq_success = self.acquire(task_from, n_slots, timeout, cancel_if)

            # If successful, push those resources to the target pool
            if acq_success:
                # Mark resources as available
                for _ in range(n_slots):
                    self._availability[task_to].release()

                # Record changes to the total pool size
                with self._allocation_lock[task_to]:
                    self._allocation[task_from] -= n_slots
                    self._allocation[task_to] += n_slots
                logger.info(f'Transferred {n_slots} slots from {task_from} to {task_to}')
                # TODO (wardlt): Eventually provide some mechanism to inform a batch
                #   system that resources allocated to ``from_task`` should be released

        return acq_success


class ReallocatorThread(Thread):
    """Thread that reallocates resources until an event is set.

    Create a thread by defining the procedure the thread should follow for reallocation
    (e.g., from where to gather resources, where to store them, where to put them when done).

    The resource allocation thread is stopped by calling ``obj.stop_event.set()``.
    Note that you can provide an Event object to the initializer to use instead of
    the ``stop_event`` attribute.

    Runs as a daemon thread."""

    def __init__(self, resource_counter: ResourceCounter,
                 gather_from: Optional[str], gather_to: Optional[str],
                 disperse_to: Optional[str], max_slots: Optional[int] = None,
                 stop_event: Optional[Event] = None,
                 slot_step: int = 1, logger_name: Optional[str] = None):
        """
        Args:
            resource_counter: Resource counter used to track resources
            stop_event: Event which controls when the thread should give resources back. If unset, a new Event is created.
            logger_name: Name of the logger, if desired
            gather_from: Name of a resource pool from which to acquire resources
            gather_to: Name of the resource pool to place re-allocated resources
            disperse_to: Name of the resource pool to move resources to after function completes
            max_slots: Maximum number of resource slots to acquire
            slot_step: Number of slots to acquire per request
        """
        super().__init__(daemon=True)

        self.resource_counter = resource_counter
        if stop_event is not None:
            self.stop_event = stop_event
        else:
            self.stop_event = Event()
        self.gather_from = gather_from
        self.gather_to = gather_to
        self.disperse_to = disperse_to
        self.max_slots = max_slots
        self.slot_step = slot_step

        self.logger = logger if logger_name is None else logging.getLogger(logger_name)

    def run(self) -> None:
        self.logger.info(f'Starting resource allocation thread. Allocating a maximum of {self.max_slots} to {self.gather_to} from'
                         f' {self.gather_from} in steps of {self.slot_step}')

        # Acquire resources until either the maximum is reached, or the event is triggered
        while (self.max_slots is None or self.resource_counter.allocated_slots(self.gather_to) < self.max_slots) \
                and not self.stop_event.is_set():
            self.resource_counter.reallocate(self.gather_from, self.gather_to,
                                             self.slot_step, cancel_if=self.stop_event)

        # Once the stop event is triggered, move the resources to a specified pool
        self.logger.info('Waiting for stop condition to be set')
        self.stop_event.wait()
        if self.gather_to != self.disperse_to:
            self.resource_counter.reallocate(self.gather_to, self.disperse_to, "all")
        self.logger.info('Resource allocation thread exiting')



"""Base class for the Task Server"""
import os
import platform

from abc import ABCMeta, abstractmethod
from multiprocessing import Process
from time import perf_counter
from typing import Optional, Callable
import logging

from colmena.exceptions import KillSignalException, TimeoutException
from colmena.redis.queue import TaskServerQueues
from colmena.models import Result, FailureInformation
from colmena.proxy import resolve_proxies_async

logger = logging.getLogger(__name__)


class BaseTaskServer(Process, metaclass=ABCMeta):
    """Abstract class for the Colmena Task Server, which manages the execution
    of different asks

    Clients submit task requests to the server by pushing them to a Redis queue,
    and then receive results from a second queue.

    Different implementations vary in how the queue is processed.

    Start the task server by first instantiating it and then calling :meth:`start`
    to launch the server in a separate process.

    The task server can be stopped by pushing a ``None`` to the task queue,
    signaling that no new tasks will be incoming. The remaining tasks will
    continue to be pushed to the output queue.
    """

    def __init__(self, queues: TaskServerQueues, timeout: Optional[int] = None):
        """
        Args:
            queues (TaskServerQueues): Queues for the task server
            timeout (int): Timeout, if desired
        """
        super().__init__()
        self.queues = queues
        self.timeout = timeout

    @abstractmethod
    def process_queue(self):
        """Evaluate a single task from the queue"""
        pass

    def listen_and_launch(self):
        logger.info('Begin pulling from task queue')
        while True:
            try:
                self.process_queue()
            except KillSignalException:
                logger.info('Kill signal received')
                return
            except TimeoutException:
                logger.info('Timeout while waiting on task queue')
                return

    @abstractmethod
    def _cleanup(self):
        """Close out any resources needed by the task server"""
        pass

    def run(self) -> None:
        """Launch the thread and start running tasks

        Blocks until the inputs queue is closed and all tasks have completed"""
        logger.info(f"Started task server {self.__class__.__name__} on {self.ident}")

        # Loop until queue has closed
        self.listen_and_launch()

        # Shutdown any needed functions
        self._cleanup()


def run_and_record_timing(func: Callable, result: Result) -> Result:
    """Run a function and also return the runtime

    Args:
        func: Function to invoke
        result: Result object describing task request
    Returns:
        Result object with the serialized result
    """
    # Mark that compute has started on the worker
    result.mark_compute_started()

    # Unpack the inputs
    result.time_deserialize_inputs = result.deserialize()

    # Start resolving any proxies in the input asynchronously
    start_time = perf_counter()
    resolve_proxies_async(result.args)
    resolve_proxies_async(result.kwargs)
    result.time_async_resolve_proxies = perf_counter() - start_time

    # Execute the function
    start_time = perf_counter()
    success = True
    try:
        output = func(*result.args, **result.kwargs)
    except BaseException as e:
        output = None
        success = False
        result.failure_info = FailureInformation.from_exception(e)
    finally:
        end_time = perf_counter()

    # Store the results
    result.set_result(output, end_time - start_time)
    if not success:
        result.success = False

    # Add the worker information into the tasks, if available
    if result.task_info is None:
        result.task_info = {}
    for tag in ['PARSL_WORKER_RANK', 'PARSL_WORKER_POOL_ID']:
        if tag in os.environ:
            result.task_info[tag] = os.environ[tag]
    result.task_info['executor'] = platform.node()

    # Re-pack the results
    result.time_serialize_results = result.serialize()

    return result



"""Task server based on FuncX

FuncX provides the ability to execute functions on remote "endpoints" that provide access to computational resources (e.g., cloud providers, HPC).
Tasks and results are communicated to/from the endpoint through a cloud service secured using Globus Auth."""

import logging
from functools import partial, update_wrapper
from typing import Dict, Callable, Optional, Tuple
from concurrent.futures import Future

from funcx import FuncXClient
from funcx.sdk.executor import FuncXExecutor

from colmena.redis.queue import TaskServerQueues
from colmena.task_server.base import BaseTaskServer, run_and_record_timing

from colmena.models import Result, FailureInformation

logger = logging.getLogger(__name__)


class FuncXTaskServer(BaseTaskServer):
    """Task server that uses FuncX to execute tasks on remote systems

    Create a FuncXTaskServer by providing a dictionary of functions along with a FuncX endpoint ID
    mapped to the `endpoint <https://funcx.readthedocs.io/en/latest/endpoints.html>`_
    on which it should run. The task server will wrap the provided function
    in an interface that tracks execution information (e.g., runtime) and
    `registers <https://funcx.readthedocs.io/en/latest/sdk.html#registering-functions>`_
    the wrapped function with FuncX.
    You must also provide a :class:`FuncXClient` that the task server can use to authenticate with the
    FuncX web service.

    The task server works using the :class:`FuncXExecutor` to communicate with FuncX via a websocket.
    `FuncXExecutor` receives completed work and we use callbacks on the Python :class:`Future` objects
    to send that completed work back to the task queue.
    """

    def __init__(self, methods: Dict[Callable, str],
                 funcx_client: FuncXClient,
                 queues: TaskServerQueues,
                 timeout: Optional[int] = None):
        """
        Args:
            methods: Map of functions to the endpoint on which it will run
            funcx_client: Authenticated FuncX client
            queues: Queues used to communicate with thinker
            timeout: Timeout for requests from the task queue
        """
        super(FuncXTaskServer, self).__init__(queues, timeout)

        # Store the FuncX client
        self.fx_client = funcx_client

        # Create a function with the latest version of the wrapper function
        self.registered_funcs: Dict[str, Tuple[Callable, str]] = {}  # Function name -> (funcX id, endpoints)
        for func, endpoint in methods.items():
            # Make a wrapped version of the function
            func_name = func.__name__
            new_func = partial(run_and_record_timing, func)
            update_wrapper(new_func, func)

            # Store the FuncX information for the function
            self.registered_funcs[func_name] = (new_func, endpoint)

        # Create the executor and queue of tasks to be submitted back to the user
        self.fx_exec = FuncXExecutor(self.fx_client)

    def _perform_callback(self, future: Future, result: Result, topic: str):
        """Send a completed result back to queue. Used as a callback for complete tasks

        Args:
            future: Future created by FuncX
            result: Initial result object. Used if the future throws an exception
            topic: Topic used to send back to the user
        """

        task_exc = future.exception()

        # If it was, send back a modified copy of the input structure
        if future.exception() is not None:
            # Mark it as unsuccessful and capture the exception information
            result.success = False
            result.failure_info = FailureInformation.from_exception(task_exc)
        else:
            # If not, the result object is the one we need
            result = future.result()

        # Put them back in the pipe with the proper topic
        self.queues.send_result(result, topic)

    def process_queue(self):
        while True:
            # Get the next task from the queue
            topic, task = self.queues.get_task(self.timeout)

            # Lookup the appropriate function ID and endpoint
            func, endp_id = self.registered_funcs[task.method]

            # Submit it to FuncX to be executed
            future: Future = self.fx_exec.submit(func, task, endpoint_id=endp_id)
            logger.info(f'Submitted {task.method} to run on {endp_id}')

            # Create the callback
            future.add_done_callback(lambda x: self._perform_callback(x, task, topic))

    def _cleanup(self):
        self.fx_exec.shutdown()



"""Exceptions type specific to the colmena"""


class KillSignalException(BaseException):
    """Server has received a signal to stop"""


class TimeoutException(BaseException):
    """Timeout on listening to a queue has occurred"""



"""Wrappers for Redis queues."""

import logging
from typing import Optional, Any, Tuple, Dict, Iterable, Union

import redis

import proxystore as ps

from colmena.exceptions import TimeoutException, KillSignalException
from colmena.models import Result, SerializationMethod

logger = logging.getLogger(__name__)


def _error_if_unconnected(f):
    def wrapper(queue: 'RedisQueue', *args, **kwargs) -> Any:
        if not queue.is_connected:
            raise ConnectionError('Not connected')
        return f(queue, *args, **kwargs)
    return wrapper


def make_queue_pairs(hostname: str, port: int = 6379, name='method',
                     serialization_method: Union[str, SerializationMethod] = SerializationMethod.JSON,
                     keep_inputs: bool = True,
                     clean_slate: bool = True,
                     topics: Optional[Iterable[str]] = None,
                     value_server_threshold: Optional[int] = None,
                     value_server_hostname: Optional[str] = None,
                     value_server_port: Optional[int] = None)\
        -> Tuple['ClientQueues', 'TaskServerQueues']:
    """Make a pair of queues for a server and client

    Args:
        hostname (str): Hostname of the Redis server
        port (int): Port on which to access Redis
        name (str): Name of the MethodServer
        serialization_method (SerializationMethod): serialization type for inputs/outputs
        keep_inputs (bool): Whether to keep the inputs after the method has finished executing
        clean_slate (bool): Whether to flush the queues before launching
        topics ([str]): List of topics used when having the client filter different types of tasks
        value_server_threshold (int): Input/output objects larger than this threshold
            (in bytes) will be stored in the value server. If None, value server
            is ignored
        value_server_hostname (str): Optional redis server hostname for value server.
            If the value server is being used but this option is not provided,
            the redis server for the task queues will be used.
        value_server_port (int): See `value_server_hostname`
    Returns:
        (ClientQueues, TaskServerQueues): Pair of communicators set to use the correct channels
    """

    return (ClientQueues(hostname, port, name, serialization_method, keep_inputs,
                         topics, value_server_threshold, value_server_hostname,
                         value_server_port),
            TaskServerQueues(hostname, port, name, topics=topics, clean_slate=clean_slate))


class RedisQueue:
    """A basic redis queue for communications used by the task server

    A queue is defined by its prefix and a "topic" designation.
    The full list of available topics is defined when creating the queue,
    and simplifies writing software that waits for only certain types
    of messages without needing to manage several "queue" objects.
    By default, the :meth:`get` methods for the queue
    listen on all topics and the :meth:`put` method pushes to the default topic.
    You can put messages into certain "topical" queues and wait for responses
    that are from a single topic.

    The queue only connects when the `connect` method is called to avoid
    issues with passing an object across processes."""

    def __init__(self, hostname: str, port: int = 6379, prefix='pipeline',
                 topics: Optional[Iterable[str]] = None):
        """
        Args:
            hostname (str): Hostname of the Redis server
            port (int): Port on which to access Redis
            prefix (str): Name of the Redis queue
            topics ([str]): List of special topics
        """
        self.hostname = hostname
        self.port = port
        self.redis_client = None
        self.prefix = prefix

        # Create the list of accepted topics
        if topics is None:
            topics = set()
        else:
            topics = set(topics)
        topics.add("default")
        assert not any('_' in t for t in topics), "Topic names may not contain underscores"
        self._all_queues = [f'{prefix}_{t}' for t in topics]

    def __setstate__(self, state):
        self.__dict__ = state

        # If you find the RedisClient placeholder,
        #  attempt to reconnect
        if self.redis_client == 'connected':
            self.redis_client = None
            self.connect()

    def __getstate__(self):
        state = self.__dict__.copy()

        # If connected, remove the unpicklable RedisClient and
        #  put a placeholder instead
        if self.is_connected:
            state['redis_client'] = 'connected'
        return state

    def connect(self):
        """Connect to the Redis server"""
        try:
            if not self.redis_client:
                self.redis_client = redis.StrictRedis(
                    host=self.hostname, port=self.port, decode_responses=True)
                self.redis_client.ping()  # Ping is needed to detect if connection failed
        except redis.exceptions.ConnectionError:
            logger.warning(f"ConnectionError while trying to connect to Redis@{self.hostname}:{self.port}")
            raise

    def disconnect(self):
        """Disconnet from the server

        Useful if sending the connection object to another process"""
        self.redis_client = None

    @_error_if_unconnected
    def get(self, timeout: int = None, topic: Optional[str] = None) -> Optional[Tuple[str, str]]:
        """Get an item from the redis queue

        Args:
            timeout (int): Timeout for the blocking get in seconds
            topic (str): Which topical queue to read from. If ``None``, wait for all topics
        Returns:
            If timeout occurs, output is ``None``. Else:
            - (str) Topic of the item
            - (str) Value from the redis queue
        """

        # Get the topic(s)
        if topic is None:
            queues = self._all_queues
        else:
            queues = f'{self.prefix}_{topic}'
            assert queues in self._all_queues, f"Unrecognized topic: {topic}"

        # Wait for the result
        try:
            if timeout is None:
                queue, result = self.redis_client.blpop(queues)
            else:
                output = self.redis_client.blpop(queues, timeout=int(timeout))
                if output is None:
                    return output
                queue, result = output

            # Strip off the prefix for the queue
            return queue.split("_")[-1], result
        except redis.exceptions.ConnectionError:
            print(f"ConnectionError while trying to connect to Redis@{self.hostname}:{self.port}")
            raise

    @_error_if_unconnected
    def put(self, input_data: str, topic: str = 'default'):
        """Push data to a Redis queue

        Args:
            input_data (str): Message to be sent
            topic (str): Topic of the queue
        """

        # Make the queue name
        queue = f'{self.prefix}_{topic}'
        assert queue in self._all_queues, f'Unrecognized topic: {topic}'

        # Send it to the task server
        try:
            assert len(input_data) < 512 * 1024 * 1024, "Value too large for Redis. Task will fail"
            output = self.redis_client.rpush(queue, input_data)
            if not isinstance(output, int):
                raise ValueError(f'Message failed to push to {queue}')
        except redis.exceptions.ConnectionError:
            logger.warning(f"ConnectionError while trying to connect to Redis@{self.hostname}:{self.port}")
            raise

    @_error_if_unconnected
    def flush(self):
        """Flush the Redis queue"""
        for queue in self._all_queues:
            try:
                self.redis_client.delete(queue)
            except AttributeError:
                raise Exception("Queue is empty/flushed")
            except redis.exceptions.ConnectionError:
                logger.warning(f"ConnectionError while trying to connect to Redis@{self.hostname}:{self.port}")
                raise

    @property
    def is_connected(self):
        return self.redis_client is not None


class ClientQueues:
    """Provides communication of method requests and results with the task server

    This queue wraps communication with the underlying Redis queue and also handles communicating
    requests using the :class:`Result` messaging format.
    Method requests are encoded in the Result format along with any options (e.g., serialization method)
    for the communication and automatically serialized.
    Results are automatically de-serialized upon receipt.

    The queue also generates stores task performance results, such as the timestamp for when messages were created
    and runtime for serialization.
    """

    def __init__(self, hostname: str, port: int = 6379, name: Optional[str] = None,
                 serialization_method: Union[str, SerializationMethod] = SerializationMethod.JSON,
                 keep_inputs: bool = True,
                 topics: Optional[Iterable] = None,
                 value_server_threshold: Optional[int] = None,
                 value_server_hostname: Optional[str] = None,
                 value_server_port: Optional[int] = None):
        """
        Args:
            hostname (str): Hostname of the Redis server
            port (int): Port on which to access Redis
            name (int): Name of the MethodServer
            serialization_method (SerializationMethod): Method used to store the input
            keep_inputs (bool): Whether to keep inputs after method results are stored
            topics ([str]): List of topics used when having the client filter different types of tasks
            value_server_threshold (int): Threshold (bytes) to store objects in
                value server. If None, the value server is not used.
            value_server_hostname (str): Optional redis server hostname for the
                value server. If not provided and the value server is used,
                defaults to `hostname`.
            value_server_port (str): Optional redis server port for the
                value server. If not provided and the value server is used,
                defaults to `port`.
        """

        # Store the result communication options
        self.serialization_method = serialization_method
        self.keep_inputs = keep_inputs
        self.value_server_threshold = value_server_threshold

        if self.value_server_threshold is not None:
            if self.serialization_method.lower() != 'pickle':
                raise ValueError('Serialization method must be pickle to use the value server')
            if value_server_hostname is None:
                value_server_hostname = hostname
            if value_server_port is None:
                value_server_port = port
            ps.store.init_store(
                ps.store.STORES.REDIS,
                name='redis',
                hostname=value_server_hostname,
                port=value_server_port
            )
            logger.debug('Initialized value server using Redis server at '
                         f'{value_server_hostname}:{value_server_port}')

        self.value_server_hostname = value_server_hostname
        self.value_server_port = value_server_port

        # Make the queues
        self.outbound = RedisQueue(hostname, port, 'inputs' if name is None else f'{name}_inputs', topics=topics)
        self.inbound = RedisQueue(hostname, port, 'results' if name is None else f'{name}_results', topics=topics)

        # Attempt to connect
        self.outbound.connect()
        self.inbound.connect()

    def send_inputs(self, *input_args: Any, method: str = None,
                    input_kwargs: Optional[Dict[str, Any]] = None,
                    keep_inputs: Optional[bool] = None,
                    topic: str = 'default',
                    task_info: Optional[Dict[str, Any]] = None):
        """Send inputs to be computed

        Args:
            *input_args (Any): Positional arguments to a function
            method (str): Name of the method to run. Optional
            input_kwargs (dict): Any keyword arguments for the function being run
            keep_inputs (bool): Whether to override the
            topic (str): Topic for the queue, which sets the topic for the result.
            task_info (dict): Any information used for task tracking
        """

        # Make fake kwargs, if needed
        if input_kwargs is None:
            input_kwargs = dict()

        # Determine whether to override the default "keep_inputs"
        _keep_inputs = self.keep_inputs
        if keep_inputs is not None:
            _keep_inputs = keep_inputs

        # Create a new Result object
        result = Result(
            (input_args, input_kwargs),
            method=method,
            keep_inputs=_keep_inputs,
            serialization_method=self.serialization_method,
            task_info=task_info,
            value_server_hostname=self.value_server_hostname,
            value_server_port=self.value_server_port,
            value_server_threshold=self.value_server_threshold
        )

        # Push the serialized value to the task server
        result.time_serialize_inputs = result.serialize()
        self.outbound.put(result.json(exclude_unset=True), topic=topic)
        logger.info(f'Client sent a {method} task with topic {topic}')

    def get_result(self, timeout: Optional[int] = None, topic: Optional[str] = None) -> Optional[Result]:
        """Get a value from the MethodServer

        Args:
            timeout (int): Timeout for waiting for a value
            topic (str): What topic of task to wait for. Set to ``None`` to pull all topics
        Returns:
            (Result) Result from a computation, or ``None`` if timeout is met
        """

        # Get a value
        output = self.inbound.get(timeout=timeout, topic=topic)
        logging.debug(f'Received value: {str(output)[:50]}')

        # If None, return because this is a timeout
        if output is None:
            return output
        topic, message = output

        # Parse the value and mark it as complete
        result_obj = Result.parse_raw(message)
        result_obj.time_deserialize_results = result_obj.deserialize()
        result_obj.mark_result_received()

        # Some logging
        logger.info(f'Client received a {result_obj.method} result with topic {topic}')

        return result_obj

    def send_kill_signal(self):
        """Send the kill signal to the task server"""
        self.outbound.put("null")


class TaskServerQueues:
    """Communication wrapper for the task server

    Handles receiving tasks
    """

    def __init__(self, hostname: str, port: int = 6379, name: Optional[str] = None,
                 clean_slate: bool = True, topics: Optional[Iterable[str]] = None):
        """
        Args:
            hostname (str): Hostname of the Redis server
            port (int): Port on which to access Redis
            name (str): Name of the MethodServer
            clean_slate (bool): Whether to flush the queues before launching
            topics ([str]): List of topics used when having the client filter different types of tasks
        """

        # Make the queues
        self.inbound = RedisQueue(hostname, port, 'inputs' if name is None else f'{name}_inputs', topics=topics)
        self.outbound = RedisQueue(hostname, port, 'results' if name is None else f'{name}_results', topics=topics)

        # Attempt to connect
        self.outbound.connect()
        self.inbound.connect()

        # Flush, if desired
        if clean_slate:
            self.outbound.flush()
            self.inbound.flush()

    def get_task(self, timeout: int = None) -> Tuple[str, Result]:
        """Get a task object

        Args:
            timeout (int): Timeout for waiting for a task
        Returns:
            - (str) Topic of the calculation. Used in defining which queue to use to send the results
            - (Result) Task description
        Raises:
            TimeoutException: If the timeout on the queue is reached
            KillSignalException: If the queue receives a kill signal
        """

        # Pull a record off of the queue
        output = self.inbound.get(timeout)

        # Return the kill signal
        if output is None:
            raise TimeoutException('Listening on task queue timed out')
        elif output[1] == "null":
            raise KillSignalException('Kill signal received on task queue')
        topic, message = output
        logger.debug(f'Received a task message with topic {topic} inbound queue')

        # Get the message
        task = Result.parse_raw(message)
        task.mark_input_received()
        return topic, task

    def send_result(self, result: Result, topic: str = 'default'):
        """Send a value to a client

        Args:
            result (Result): Result object to communicate back
            topic (str): Topic of the calculation
        """
        result.mark_result_sent()
        self.outbound.put(result.json(), topic=topic)



"""Base classes for 'thinking' applications that respond to tasks completing"""
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import partial, update_wrapper
from threading import Event, local, Thread
from traceback import TracebackException
from typing import Optional, Callable, List, Union

import os

import logging

from colmena.redis.queue import ClientQueues
from colmena.thinker.resources import ResourceCounter, ReallocatorThread

logger = logging.getLogger(__name__)

_DONE_REACTION_TIME = 1


def agent(func: Optional[Callable] = None, critical: bool = True):
    """Decorator that denotes a function as an "agent" thread that is launched when a Thinker process is started

    Args:
        func: Do not directly pass this variable. It is used as an argument to the decorator
        critical: Whether the "done" flag should be set once this thread finishes
    """
    def decorator(f: Callable):
        f._colmena_agent = True
        f._colmena_critical = critical
        return f
    if func is None:
        return decorator
    return decorator(func)


def _result_event_agent(thinker: 'BaseThinker', process_func: Callable, topic: Optional[str]):
    """Wrapper function for result processing agents"""
    # Wait until we get a result
    while not thinker.done.is_set():
        result = thinker.queues.get_result(timeout=_DONE_REACTION_TIME, topic=topic)
        if result is not None:
            process_func(thinker, result)


def result_processor(func: Optional[Callable] = None, topic: Optional[str] = None):
    """Decorator that builds agents which respond to results becoming available in a queue

    Decorated functions must take a single argument: a result object

    Args:
        func: Do not directly pass this variable. It is used as an argument to the decorator
        topic: Topic of the queue to pull from
    """

    def decorator(f: Callable):
        output = partial(_result_event_agent, process_func=f, topic=topic)
        output = agent(output)
        return update_wrapper(output, f)
    if func is None:
        return decorator
    return decorator(func)


def _task_submitter_agent(thinker: 'BaseThinker', process_func: Callable, task_type: str, n_slots: Union[int, str]):
    """Wrapper function for task submission agents"""
    # Determine the number of threads
    if not isinstance(n_slots, int):
        n_slots = getattr(thinker, n_slots)

    while not thinker.done.is_set():
        # Wait until resources are free or thinker.done is set
        acq_success = thinker.rec.acquire(task_type, n_slots, cancel_if=thinker.done)
        if acq_success:
            process_func(thinker)


def task_submitter(func: Optional[Callable] = None, task_type: str = None, n_slots: Union[int, str] = 1):
    """Decorator that builds agents which respond to computing resources becoming available

    Decorated functions should assume that resources are available and reserved when the function is called

    Args:
        func: Do not directly pass this variable. It is used as an argument to the decorator
        task_type: Name of task pool from which to request resources
        n_slots: Number of resources to request. Must be either a integer or the name of a class attribute
    """

    def decorator(f: Callable):
        output = partial(_task_submitter_agent, process_func=f, task_type=task_type, n_slots=n_slots)
        output = agent(output)
        return update_wrapper(output, f)
    if func is None:
        return decorator
    return decorator(func)


def _event_responder_agent(thinker: 'BaseThinker', process_func: Callable, event_name: str,
                           reallocate_resources: bool, gather_from: Optional[str],
                           gather_to: Optional[str], disperse_to: Optional[str],
                           max_slots: Union[int, str, None], slot_step: int):
    """Wrapper for event processing agents"""

    # Get the event
    if not hasattr(thinker, event_name):
        raise ValueError(f'Thinker lacks an event named {event_name}')
    event: Event = getattr(thinker, event_name)

    # Get the max_slots if set to a class attribute
    if not isinstance(max_slots, int) and max_slots is not None:
        max_slots = getattr(thinker, max_slots)

    # Loop until the thinker is completed
    reallocator_thread: Optional[ReallocatorThread] = None
    while not thinker.done.is_set():
        if event.wait(_DONE_REACTION_TIME):
            # If desired, launch the resource-allocation thread
            if reallocate_resources:
                reallocator_thread = ReallocatorThread(
                    thinker.rec, gather_from=gather_from, gather_to=gather_to,
                    disperse_to=disperse_to, max_slots=max_slots,
                    slot_step=slot_step, logger_name=thinker.logger.name + ".allocate"
                )
                reallocator_thread.start()

            # Launch the function
            process_func(thinker)

            # If we are using resource re-allocation, set the stop condition and wait for resources to be freed
            if reallocator_thread is not None:
                reallocator_thread.stop_event.set()
                reallocator_thread.join()


def event_responder(func: Optional[Callable] = None, event_name: str = None,
                    reallocate_resources: bool = False,
                    gather_from: Optional[str] = None, gather_to: Optional[str] = None,
                    disperse_to: Optional[str] = None, max_slots: Union[int, str, None] = None,
                    slot_step: int = 1):
    """Decorator that builds agents which respond to an event being set.

    The event responder can launch a thread to acquire resource temporarily.
    The thread is created if you set ``reallocate_resources=True`` in the decorator
    and transfers resources to a specific pool until the decorated function completes
    or a user-defined resource cap is set.
    You must configure from where these resources are acquired, in which resource pool
    they are placed, and where they are re-allocated after the thread completes.
    The thread will allocate up to the maximum number of slots defined and
    then reallocate _all slots available to that pool_ to the designated resource.


    The Thinker associated with this agent must have a class attribute that is an :class:`Event`
    with the same name as ``event_name``.

    Args:
        func: Do not directly pass this variable. It is used as an argument to the decorator
        event_name: Name of the event to watch
        reallocate_resources: Whether to re-allocate resources while function is running
        gather_from: Name of a resource pool from which to acquire resources
        gather_to: Name of the resource pool to place re-allocated resources
        disperse_to: Name of the resource pool to move resources to after function completes
        max_slots: Maximum number of resource slots to acquire. Can be an integer,
            the name of a class attribute of the thinker, or 'none' it no maximum is needed
        slot_step: Number of slots to acquire per request
    """

    # Make sure the re-allocation logic is configured reasonably
    if reallocate_resources and gather_to == gather_from:
        raise ValueError('Resources should be drawn and stored in different pools')

    def decorator(f: Callable):
        output = partial(_event_responder_agent, process_func=f, event_name=event_name,
                         reallocate_resources=reallocate_resources, gather_from=gather_from, gather_to=gather_to,
                         disperse_to=disperse_to, max_slots=max_slots, slot_step=slot_step)
        output = agent(output)
        return update_wrapper(output, f)
    if func is None:
        return decorator
    return decorator(func)


def _launch_agent(func: Callable, thinker: 'BaseThinker'):
    """Shim function for launching an agent

    Sets the thread-local variables for a class, such as its name and default topic
    """

    # Set the thread-local options for this agent
    name = func.__name__
    thinker.local_details.name = name
    thinker.local_details.logger = thinker.make_logger(name)

    # Mark that this thread has launched
    thinker.logger.info(f'{name} started')

    # Launch it
    try:
        func(thinker)
    finally:
        # If a "critical" function, set the "done" flag
        if getattr(func, '_colmena_critical', False):
            thinker.done.set()

        # Mark that the thread has crashed
        thinker.logger.info(f'{name} completed')


class _AgentData(local):
    """Data local to a certain agent thread

    Attributes:
        logger: Logger for this thread
        name (str): Name of the thread
    """

    def __init__(self, logger: logging.Logger):
        """
        Args:
            logger: Logger to use for this thread
        """
        self.logger = logger
        self.name: Optional[str] = None


class BaseThinker(Thread):
    """Base class for dataflow program that steers a Colmena application

    The intent of this class is to simplify writing an dataflow programs using Colmena.
    When implementing a subclass, write each operation in the program as class method.
    Each method should take no inputs and produce no output, and could be thought of as
    an "operation" or "agent" that will run as a thread.

    Each agent communicates with others via `queues <https://docs.python.org/3/library/queue.html>`_
    or other `threading objects <https://docs.python.org/3/library/threading.html#>`_ and
    the Colmena task server via the :class:`ClientQueues`.
    The only communication method available by default is a class attribute named ``done``
    that is used to signal that the program should terminate.

    Denote each of these agents with the :meth:`agent` decorator, as in:

    .. code-block: python

        class ExampleThinker(BaseThinker):

            @agent
            def function(self):
                return True

    The decorator will tell Colmena to launch that method as a separate thread
    when the "Thinker" thread is started.
    Colmena will also create a distinct logger for each of the agents to that is
    accessible as the :meth:`logger` property.

    Start the thinker by calling ``.start()``
    """

    def __init__(self, queue: ClientQueues, resource_counter: Optional[ResourceCounter] = None,
                 daemon: bool = True, **kwargs):
        """
            Args:
                queue: Queue wrapper used to communicate with task server
                resource_counter: Utility to used track resource utilization
                daemon: Whether to launch this as a daemon thread
                **kwargs: Options passed to :class:`Thread`
        """
        super().__init__(daemon=daemon, **kwargs)

        # Define thinker-wide collectives
        self.rec = resource_counter
        self.queues = queue

        # Create some basic events and locks
        self.done: Event = Event()

        # Thread-local stuff, like the default queue and name
        self.local_details = _AgentData(self.make_logger())

    @property
    def logger(self) -> logging.Logger:
        """Get the logger for the active thread"""
        return self.local_details.logger

    def make_logging_handler(self) -> Optional[logging.Handler]:
        """Override to create a distinct logging handler for log messages emitted from this object"""
        return None

    def make_logger(self, name: Optional[str] = None):
        """Make a sub-logger for our application

        Args:
            name: Name to use for the sub-logger

        Returns:
            Logger with an appropriate name
        """
        # Create the logger
        my_name = self.__class__.__name__.lower()
        if name is not None:
            my_name += "." + name
        new_logger = logging.getLogger(my_name)

        # Assign the handler to the root logger
        if name is None:
            hnd = self.make_logging_handler()
            if hnd is not None:
                new_logger.addHandler(hnd)
        return new_logger

    @classmethod
    def list_agents(cls) -> List[Callable]:
        """List all functions that map to operations within the thinker application

        Returns:
            List of methods that define agent threads
        """
        agents = []
        for n in dir(cls):
            o = getattr(cls, n)
            if hasattr(o, '_colmena_agent'):
                agents.append(o)
        return agents

    def run(self):
        """Launch all operation threads and wait until all complete

        Sets the ``done`` flag when a thread completes, then waits for all other flags to finish.

        Does not raise exceptions if a thread exits with an exception. Exception and traceback information
        are printed using logging at the ``WARNING`` level.
        """
        self.logger.info(f"{self.__class__.__name__} started. Process id: {os.getpid()}")

        threads = []
        functions = self.list_agents()
        with ThreadPoolExecutor(max_workers=len(functions)) as executor:
            # Submit all of the worker threads
            for f in functions:
                threads.append(executor.submit(_launch_agent, f, self))
            self.logger.info(f'Launched all {len(functions)} functions')

            # Wait until any one completes, then set the "gen_done" event to
            #  signal all remaining threads to finish after completing their work
            for finished in as_completed(threads):
                exc = finished.exception()
                if exc is None:
                    self.logger.info('Thread completed without problems')
                else:
                    tb = TracebackException.from_exception(exc)
                    self.logger.warning(f'Thread failed: {exc}.\nTraceback: {"".join(tb.format())}')

        self.logger.info(f"{self.__class__.__name__} completed")



"""Utilities for interacting with ProxyStore"""
import proxystore as ps

from typing import Any, Optional, Union

from colmena.models import SerializationMethod


class ColmenaSerializationFactory(ps.store.redis.RedisFactory):
    """Custom Factory for using Colmena serialization utilities"""
    def __init__(self,
                 key: str,
                 name: str,
                 hostname: str,
                 port: int,
                 serialization_method: Union[str, SerializationMethod] = SerializationMethod.PICKLE,
                 **kwargs) -> None:
        """Init ColmenaSerialization Factory

        Args:
            key (str): key corresponding to object in Redis.
            name (str): name of store to retrive objects from.
            hostname (str): hostname of Redis server containing object.
            port (int): port of Redis server containing object.
            serialization_method (str): Colmena serialization method to use
                for deserializing the object when resolved from Redis.
            kwargs: keyword arguments to pass to the RedisFactory.
        """
        self.serialization_method = serialization_method
        self.kwargs = kwargs
        super(ColmenaSerializationFactory, self).__init__(
            key, name, hostname, port, **kwargs
        )

    def __getnewargs_ex__(self):
        """Helper method for pickling

        Note:
            We override default pickling behavior because a Factory may contain
            a Future if it is being asynchronously resolved and Futures cannot
            be pickled.
        """
        return (self.key, self.name, self.hostname, self.port), {
            'serialization_method': self.serialization_method,
            **self.kwargs
        }

    def resolve(self) -> Any:
        obj_str = super(ColmenaSerializationFactory, self).resolve()
        return SerializationMethod.deserialize(self.serialization_method, obj_str)


def proxy(obj: Any,
          key: Optional[str] = None,
          is_serialized: bool = False,
          serialization_method: Union[str, SerializationMethod] = SerializationMethod.PICKLE,
          **kwargs) -> ps.proxy.Proxy:
    """Place object in Value Server and return Proxy

    Args:
        obj: object to be placed in Value Server and proxied.
        key (str): optional key to associate with object. By default, ProxyStore
            will create a key for the object (default: None).
        is_serialized (bool): True if obj is already serialized (default: False).
        serialization_method (str): serialization method to use for the object
            (default: SerializationMethod.PICKLE).
        kwargs (dict): keyword arguments to pass to ProxyStore.store.redis.RedisStore.proxy().

    Returns:
        ps.proxy.Proxy
    """
    store = ps.store.get_store('redis')
    if not is_serialized:
        obj = SerializationMethod.serialize(serialization_method, obj)
    return store.proxy(
        obj,
        key,
        serialize=False,  # Do not use ProxyStore serialization utilities
        serialization_method=serialization_method,
        factory=ColmenaSerializationFactory,
        **kwargs
    )


def resolve_proxies_async(args: Union[object, list, tuple, dict]) -> None:
    """Begin asynchronously resolving all proxies in input

    Scan inputs for instances of `Proxy` and begin asynchronously resolving.
    This is useful if you have one or more proxies that will be needed soon
    so the underlying objects can be asynchronously resolved to reduce the
    cost of the first access to the proxy.

    Args:
        args (object, list, tuple, dict): possible object or
            iterable of objects that may be ObjectProxy instances
    """
    def resolve_async_if_proxy(obj: Any) -> None:
        if isinstance(obj, ps.proxy.Proxy):
            ps.proxy.resolve_async(obj)

    if isinstance(args, ps.proxy.Proxy):
        resolve_async_if_proxy(args)
    elif isinstance(args, list) or isinstance(args, tuple):
        for x in args:
            resolve_async_if_proxy(x)
    elif isinstance(args, dict):
        for x in args:
            resolve_async_if_proxy(args[x])



import os
from setuptools import setup, find_packages

version_ns = {}
with open(os.path.join("colmena", "version.py")) as f:
    exec(f.read(), version_ns)
version = version_ns['VERSION']

with open('requirements.txt') as f:
    install_requires = f.readlines()
    
with open('README.md') as f:
    long_desc = f.read()

setup(
    name='colmena',
    version=version,
    packages=find_packages(),
    include_package_data=True,
    description='colmena: Intelligent Steerable Pipelines on HPC',
    long_description=long_desc,
    long_description_content_type='text/markdown',
    install_requires=install_requires,
    python_requires=">=3.6.*",
    extras_require={
        'funcx': ['funcx>=0.3.4', 'globus_sdk<3']  # TODO (wardlt): Remove Globus SDK version pinning once FuncX supports v3
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering"
    ],
    keywords=[
        "parsl",
        "HPC",
    ],
    author="Globus Labs",
    author_email='labs@globus.org',
    license="Apache License, Version 2.0",
    url="https://github.com/exalearn/colmena"
)



import json
import logging
import pickle as pkl
import sys
from datetime import datetime
from enum import Enum
from time import perf_counter
from traceback import TracebackException
from typing import Any, Tuple, Dict, Optional, Union

from pydantic import BaseModel, Field

import colmena
import proxystore as ps

logger = logging.getLogger(__name__)


# TODO (wardlt): Merge with FuncX's approach?
class SerializationMethod(str, Enum):
    """Serialization options"""

    JSON = "json"  # Serialize using JSON
    PICKLE = "pickle"  # Pickle serialization

    @staticmethod
    def serialize(method: 'SerializationMethod', data: Any) -> str:
        """Serialize an object using a specified method

        Args:
            method: Method used to serialize the object
            data: Object to be serialized
        Returns:
            Serialized data
        """

        if method == "json":
            return json.dumps(data)
        elif method == "pickle":
            return pkl.dumps(data).hex()
        else:
            raise NotImplementedError(f'Method {method} not yet implemented')

    @staticmethod
    def deserialize(method: 'SerializationMethod', message: str) -> Any:
        """Deserialize an object

        Args:
            method: Method used to serialize the message
            message: Message to deserialize
        Returns:
            Result object
        """

        if method == "json":
            return json.loads(message)
        elif method == "pickle":
            return pkl.loads(bytes.fromhex(message))
        else:
            raise NotImplementedError(f'Method {method} not yet implemented')


class FailureInformation(BaseModel):
    """Stores information about a task failure"""

    exception: str = Field(..., description="The exception returned by the failed task")
    traceback: Optional[str] = Field(None, description="Full stack trace for exception, if available")

    @classmethod
    def from_exception(cls, exc: BaseException) -> 'FailureInformation':
        tb = TracebackException.from_exception(exc)
        return cls(exception=repr(exc), traceback="".join(tb.format()))


class Result(BaseModel):
    """A class which describes the inputs and results of the calculations evaluated by the MethodServer

    Each instance of this class stores the inputs and outputs to the function along with some tracking
    information allowing for performance analysis (e.g., time submitted to Queue, time received by client).
    All times are listed as Unix timestamps.

    The Result class also handles serialization of the data to be transmitted over a RedisQueue
    """

    # Core result information
    inputs: Union[Tuple[Tuple[Any, ...], Dict[str, Any]], str] = \
        Field(None, description="Input to a function. Positional and keyword arguments. The `str` data type "
                                "is for internal use and is used when communicating serialized objects.")
    value: Any = Field(None, description="Output of a function")
    method: Optional[str] = Field(None, description="Name of the method to run.")
    success: Optional[bool] = Field(None, description="Whether the task completed successfully")

    # Store task information
    task_info: Optional[Dict[str, Any]] = Field(default_factory=dict, description="Task tracking information to be transmitted "
                                                                                  "along with inputs and results. User provided")
    failure_info: Optional[FailureInformation] = Field(None, description="Messages about task failure. Provided by Task Server")

    # Performance tracking
    time_created: float = Field(None, description="Time this value object was created")
    time_input_received: float = Field(None, description="Time the inputs was received by the task server")
    time_compute_started: float = Field(None, description="Time workflow process began executing a task")
    time_result_sent: float = Field(None, description="Time message was sent from the server")
    time_result_received: float = Field(None, description="Time value was received by client")

    time_running: float = Field(None, description="Runtime of the method, if available")
    time_serialize_inputs: float = Field(None, description="Time required to serialize inputs on client")
    time_deserialize_inputs: float = Field(None, description="Time required to deserialize inputs on worker")
    time_serialize_results: float = Field(None, description="Time required to serialize results on worker")
    time_deserialize_results: float = Field(None, description="Time required to deserialize results on client")
    time_async_resolve_proxies: float = Field(None, description="Time required to scan function inputs and start async resolves of proxies")

    # Serialization options
    serialization_method: SerializationMethod = Field(SerializationMethod.JSON,
                                                      description="Method used to serialize input data")
    keep_inputs: bool = Field(True, description="Whether to keep the inputs with the result object or delete "
                                                "them after the method has completed")
    value_server_hostname: Optional[str] = Field(None, description="Value server hostname")
    value_server_port: Optional[str] = Field(None, description="Value server port")
    value_server_threshold: int = Field(
        None, description="Object size threshold (bytes) at which input/value "
                          "objects are stored in value server before serialization")

    def __init__(self, inputs: Tuple[Tuple[Any], Dict[str, Any]], **kwargs):
        """
        Args:
             inputs (Any, Dict): Inputs to a function. Separated into positional and keyword arguments
        """
        super().__init__(inputs=inputs, **kwargs)

        # Mark "created" only if the value is not already set
        if 'time_created' not in kwargs:
            self.time_created = datetime.now().timestamp()

    @property
    def args(self) -> Tuple[Any]:
        return tuple(self.inputs[0])

    @property
    def kwargs(self) -> Dict[str, Any]:
        return self.inputs[1]

    def mark_result_received(self):
        """Mark that a completed computation was received by a client"""
        self.time_result_received = datetime.now().timestamp()

    def mark_input_received(self):
        """Mark that a task server has received a value"""
        self.time_input_received = datetime.now().timestamp()

    def mark_compute_started(self):
        """Mark that the compute for a method has started"""
        self.time_compute_started = datetime.now().timestamp()

    def mark_result_sent(self):
        """Mark when a result is sent from the task server"""
        self.time_result_sent = datetime.now().timestamp()

    def set_result(self, result: Any, runtime: float = None):
        """Set the value of this computation

        Automatically sets the "time_result_completed" field and, if known, defines the runtime.

        Will delete the inputs to the function if the user specifies ``self.return_inputs == False``.
        Removing the inputs once the result is known can save communication time

        Args:
            result: Result to be stored
            runtime (float): Runtime for the function
        """
        self.value = result
        if not self.keep_inputs:
            self.inputs = ((), {})
        self.time_running = runtime
        self.success = True

    def serialize(self) -> float:
        """Stores the input and value fields as a pickled objects

        Returns:
            (float) Time to serialize
        """
        start_time = perf_counter()
        _value = self.value
        _inputs = self.inputs

        def _serialize_and_proxy(value, evict=False):
            """Helper function for serializing and proxying"""
            # Serialized object before proxying to compare size of serialized
            # object to value server threshold
            value_str = SerializationMethod.serialize(
                self.serialization_method, value
            )

            if (
                    self.value_server_threshold is not None and
                    sys.getsizeof(value_str) >= self.value_server_threshold and
                    not isinstance(value, ps.proxy.Proxy)
            ):
                # Proxy the value. Note: we use the id of the object as the key
                # so calling proxy() on the same object multiple times
                # does not create multiple copies in the value server.
                value_proxy = colmena.proxy.proxy(
                    value_str,
                    key=str(id(value)),
                    is_serialized=True,
                    serialization_method=self.serialization_method,
                    evict=evict
                )
                logger.debug(f'Proxied object of type {type(value)} with id={id(value)}')
                # Serialize the proxy. This is efficient since the proxy is
                # just a reference + metadata about the value
                value_str = SerializationMethod.serialize(
                    self.serialization_method, value_proxy
                )

            return value_str

        try:
            # Each value in *args and **kwargs is serialized independently
            args = tuple(map(_serialize_and_proxy, _inputs[0]))
            kwargs = {k: _serialize_and_proxy(v) for k, v in _inputs[1].items()}
            self.inputs = (args, kwargs)

            # The entire result is serialized as one object. Pass evict=True
            # so the value is evicted from the value server once it is resolved
            # by the thinker.
            if _value is not None:
                self.value = _serialize_and_proxy(_value, evict=True)

            return perf_counter() - start_time
        except Exception as e:
            # Put the original values back
            self.inputs = _inputs
            self.value = _value
            raise e

    def deserialize(self) -> float:
        """De-serialize the input and value fields

        Returns:
            (float) The time required to deserialize
        """
        # Check that the data is actually a string
        start_time = perf_counter()
        _value = self.value
        _inputs = self.inputs

        def _deserialize(value):
            if not isinstance(value, str):
                return value
            return SerializationMethod.deserialize(self.serialization_method, value)

        if isinstance(_inputs, str):
            _inputs = SerializationMethod.deserialize(self.serialization_method, _inputs)

        try:
            # Deserialize each value in *args and **kwargs
            args = tuple(map(_deserialize, _inputs[0]))
            kwargs = {k: _deserialize(v) for k, v in _inputs[1].items()}
            self.inputs = (args, kwargs)

            # Deserialize result if it exists
            if _value is not None:
                self.value = _deserialize(_value)

            return perf_counter() - start_time
        except Exception as e:
            # Put the original values back
            self.inputs = _inputs
            self.value = _value
            raise e



"""Parsl task server and related utilities"""
import os
import logging
from functools import partial
from queue import Queue
from threading import Thread
from concurrent.futures import wait, Future
from time import sleep
from typing import Optional, List, Callable, Tuple, Dict, Union, Set

import parsl
from parsl import python_app, ThreadPoolExecutor
from parsl.config import Config
from parsl.app.python import PythonApp
from parsl.dataflow.futures import AppFuture

from colmena.models import Result, FailureInformation
from colmena.task_server.base import BaseTaskServer
from colmena.redis.queue import TaskServerQueues
from colmena.task_server.base import run_and_record_timing

logger = logging.getLogger(__name__)


@python_app(executors=['_output_workers'])
def output_result(queues: TaskServerQueues, topic: str, result_obj: Result):
    """Submit the function result to the Redis queue

    Args:
        queues: Queues used to communicate with Redis
        topic: Topic to assign in output queue
        result_obj: Result object containing the inputs, to be sent back with outputs
    """
    return queues.send_result(result_obj, topic=topic)


class _ErrorHandler(Thread):
    """Keeps track of the Parsl futures and reports back errors"""

    def __init__(self, future_queue: Queue, timeout: float = 5):
        """
        Args:
            future_queue (Queue): A queue on which to receive
            timeout (float): How long to wait before checking for new futures
        """
        super().__init__(daemon=True, name='error_handler')
        self.future_queue = future_queue
        self.timeout = timeout
        self.kill = False

    def run(self) -> None:
        # Initialize the list of futures
        logger.info('Starting the error-handler thread')
        futures: Set[Future] = set()

        # Continually look for new jobs
        while not self.kill:
            # Pull from the queue until empty
            #  This operation assumes we have only one thread reading from the queue
            #  Otherwise, the queue could become empty between checking status and then pulling
            while not self.future_queue.empty():
                futures.add(self.future_queue.get())

            # If no futures, wait for the timeout
            if len(futures) == 0:
                sleep(self.timeout)
            else:
                done, not_done = wait(futures, timeout=self.timeout)

                # If there are entries that are complete, check if they have errors
                for task in done:
                    # Check if an exception was raised
                    exc = task.exception()
                    if exc is None:
                        logger.debug(f'Task completed: {task}')
                        continue
                    logger.warning(f'Task {task} with an exception: {exc}')

                    # Pull out the result objects
                    queues: TaskServerQueues = task.task_def['args'][0]
                    topic: str = task.task_def['args'][1]
                    method_task = task.task_def['depends'][0]
                    task_exc = method_task.exception()
                    result_obj: Result = method_task.task_def['args'][0]

                    # Mark it as unsuccessful and capture the exception information
                    result_obj.success = False
                    result_obj.failure_info = FailureInformation.from_exception(task_exc)

                    # Send it to the client
                    queues.send_result(result_obj, topic=topic)

                # Display run information
                if len(done) > 0:
                    logger.debug(f'Cleared {len(done)} futures from Parsl task queue')

                # Loop through the incomplete tasks
                futures = not_done


class ParslTaskServer(BaseTaskServer):
    """Task server based on Parsl

    Create a Parsl task server by first creating a resource configuration following
    the recommendations in `the Parsl documentation
    <https://parsl.readthedocs.io/en/stable/userguide/configuring.html>`_.
    Then instantiate a task server with a list of Python functions,
    configurations defining on which Parsl executors each function can run,
    and the Parsl resource configuration.
    The executor(s) for each function can be defined with a combination
    of per method specifications

    .. code-block:: python

        ParslTaskServer([(f, {'executors': ['a']})], queues, config)

    and also using a default executor

    .. code-block:: python

        ParslTaskServer([f], queues, config, default_executors=['a'])

    Further configuration options for each method can be defined
    in the list of methods.

    **Technical Details**

    The task server stores each of the supplied methods as Parsl "PythonApp" classes.
    Tasks are launched using these PythonApps after being received on the queue.
    The Future provided when requesting the method invocation is then passed
    to second PythonApp that pushes the result of the function to the output
    queue after it completes.
    That second, "output_result," function runs on threads of the same
    process as this task server.
    There is also a separate thread that monitors for Futures that yield an error
    before the "output_result" function and sends back the error messages.
    """

    def __init__(self, methods: List[Union[Callable, Tuple[Callable, Dict]]],
                 queues: TaskServerQueues,
                 config: Config,
                 timeout: Optional[int] = None,
                 default_executors: Union[str, List[str]] = 'all',
                 num_output_workers: int = 4):
        """

        Args:
            methods (list): List of methods to be served.
                Each element in the list is either a function or a tuple where the first element
                is a function and the second is a dictionary of the arguments being used to create
                the Parsl ParslApp see `Parsl documentation
                <https://parsl.readthedocs.io/en/stable/stubs/parsl.app.app.python_app.html#parsl.app.app.python_app>`_.
            queues (TaskServerQueues): Queues for the task server
            config: Parsl configuration
            timeout (int): Timeout, if desired
            default_executors: Executor or list of executors to use by default.
            num_output_workers: Maximum number of output workers push results to the Redis queue. Will not create
                more workers than CPUs
        """
        super().__init__(queues, timeout)

        # Insert _output_workers to the thread count
        executors = config.executors.copy()
        executors.append(ThreadPoolExecutor(label='_output_workers', max_threads=num_output_workers))
        config.executors = executors

        # Get a list of default executors that _does not_ include the output workers
        if default_executors == 'all':
            default_executors = [e.label for e in executors if e.label != "_output_workers"]

        # Store the Parsl configuration
        self.config = config

        # Assemble the list of methods
        self.methods_ = {}
        for method in methods:
            # Get the options or use the defaults
            if isinstance(method, (tuple, list)):
                if len(method) != 2:
                    raise ValueError('Method description should a tuple of length 2')
                function, options = method
            else:
                function = method
                options = {'executors': default_executors}
                logger.info(f'Using default executors for {function.__name__}: {default_executors}')

            # Make the Parsl app
            name = function.__name__

            # Wrap the function in the timer class
            wrapped_function = partial(run_and_record_timing, function)
            app = PythonApp(wrapped_function, **options)

            # Store it
            self.methods_[name] = app
        logger.info(f'Defined {len(self.methods_)} methods: {", ".join(self.methods_.keys())}')

        # If only one method, store a default method
        self.default_method_ = list(self.methods_.keys())[0] if len(self.methods_) == 1 else None
        if self.default_method_ is not None:
            logger.info(f'There is only one method, so we are using {self.default_method_} as a default')

        # Create a thread to check if tasks completed successfully
        self.task_queue: Optional[Queue] = None
        self.error_checker: Optional[_ErrorHandler] = None

    def process_queue(self):
        """Evaluate a single task from the queue"""

        # Get a result from the queue
        topic, result = self.queues.get_task(self.timeout)
        logger.info(f'Received request for {result.method} with topic {topic}')

        # Determine which method to run
        if self.default_method_ and result.method is None:
            method = self.default_method_
        else:
            method = result.method

        # Submit the application
        future = self.submit_application(method, result)
        # TODO (wardlt): Implement "resubmit if task returns a new future."
        #  Requires waiting on two streams: input_queue and the queues

        # Pass the future of that operation to the output queue
        result_future = output_result(self.queues, topic, future)
        logger.debug('Pushed task to Parsl')

        # Pass the task to the "error handler"
        self.task_queue.put(result_future)

    def submit_application(self, method_name: str, result: Result) -> AppFuture:
        """Submit an application to run via Parsl

        Args:
            method_name (str): Name of the method to invoke
            result: Task description
        """
        return self.methods_[method_name](result)

    def _cleanup(self):
        """Close out any resources needed by the task server"""
        # Wait until all tasks have finished
        dfk = parsl.dfk()
        dfk.wait_for_current_tasks()
        logger.info(f"All tasks have completed for {self.__class__.__name__} on {self.ident}")

        logger.info("Shutting down the error handling thread")
        self.error_checker.kill = True

    def run(self) -> None:
        # Launch the Parsl workflow engine
        parsl.load(self.config)
        logger.info(f"Launched Parsl DFK. Process id: {os.getpid()}")

        # Create the error checker
        self.task_queue = Queue()
        self.error_checker = _ErrorHandler(self.task_queue)
        self.error_checker.start()

        # Start the loop
        super().run()
